% ****************************************************************
%   File:refine.m
%   Author: Kingsley Sage, University of Sussex, Brighton, UK
%   Description: MATLAB function to calculate independent
%                mean and covariance parameters sets for
%                each variable length state in a VLMM
%   Version: Release version 1.0
%
%   This code is supplied with no warranty whatsoever and was developed
%   with support from the EU ActIPret project and in part fulfilment of 
%   Kingsley Sage's DPhil work on 'Task Based Visual Control'. Any comments
%   or suggestions should be e-mailed to khs20@sussex.ac.uk.
%
%   The code may be reproduced freely for non-commercial use provided
%   that re-use acknowledges Kingsley Sage.
%   ***************************************************************/

function ModelOut = refine(ModelIn,C,R,Y)
% calculates an independent mean value for each variable length state in ModelIn.

if ~iscell(Y) Y={Y}; end;
M = length(ModelIn); % Number of variable states
seqs = length(Y); % Number of examplars in training set
GammaSet = {};

for i=1:M
  qq=ModelIn{i}.node;
  StateIndex(i,1)=qq(1,size(qq,2)); % The terminal symbol for ModelIn{i}.node
end;

for i=1:seqs
  GammaSet{i} = maketrellis(ModelIn,M,C,R,StateIndex,Y{i});
end;

% Now estimate mu values for each state in M ...
muM=zeros(size(C,1),M);
gammatot=zeros(M,1);
for i=1:seqs
  gammatot=gammatot+(sum(GammaSet{i},2));
  for j=1:size(Y{i},2)
    for k=1:M
      muM(:,k)=muM(:,k)+Y{i}(:,j)*GammaSet{i}(k,j);
    end;
  end;
end;
muM=muM./(gammatot*ones(1,size(C,1)))';

% Now put these mu values into the refined model ...
ModelOut={};
for i=1:M
  ModelOut{i} = ModelIn{i};
  ModelOut{i}.mu = muM(:,i);
end;

% Now estimate independent covariance sets for each state in M ...
covM=zeros(size(C,1),size(C,1),M);
for i=1:seqs
  for j=1:size(Y{i},2)
    for k=1:M
      dd = Y{i}(:,j)-muM(:,k);
      covM(:,:,k)=covM(:,:,k)+(dd*GammaSet{i}(k,j))*dd';
      % THINK!!: This is diff^2*freq which is OK ...
    end;
  end;
end;

% And put these values into the refined model ...
for k=1:M
  covM(:,:,k)=covM(:,:,k)/gammatot(k,1);
  ModelOut{k}.cov=covM(:,:,k);
end;

function gamma=maketrellis(ModelIn,M,C,R,StateIndex,Y)
 
T = size(Y,2);
LengthSigma = size(C,2); % The number of hidden states in the original Markov Model
alpha = zeros(M,T);
beta = zeros(M,T);
rho = zeros(1,T); % Scaling factors
N = size(C,1); % I.e. number of feature dimensions;
bb = zeros(M,T);

invR = inv(R);
constant = ((2*pi)^-(N/2))*sqrt(det(invR));
for (t=1:T)
  for i=1:M
    j = StateIndex(i,1); % i.e. which C and R cluster VLMM state i relates to ...
    bb(i,t)=constant * exp(-0.5*((Y(:,t)-C(:,j))'*invR*(Y(:,t)-C(:,j))));
  end;
end;

alpha(:,1) = ones(M,1) * (1/M); % Uniform prior
alpha(:,1) = alpha(:,1).*bb(:,1);
alpha(:,1) = alpha(:,1) / sum(alpha(:,1));

% And over the rest of the training sequence ...
for t=2:T
  for i1=1:M
    for i3=1:LengthSigma
      if (ModelIn{i1}.PTable(i3,2)~=-1)
        alpha(ModelIn{i1}.PTable(i3,2),t)=alpha(ModelIn{i1}.PTable(i3,2),t) + ...
                         (ModelIn{i1}.PTable(i3,1)*alpha(ModelIn{i1}.ref,t-1));
      end;
    end;
  end;
  alpha(:,t) = alpha(:,t).*bb(:,t);
  rho(t) = sum(alpha(:,t));
  alpha(:,t) = alpha(:,t) / rho(t);
end;

beta(:,T) = ones(M,1);
for t=T-1:-1:1
  for i1=1:M
    for i3=1:LengthSigma
      if (ModelIn{i1}.PTable(i3,2)~=-1)
        beta(i1,t)=beta(i1,t)+ ...
           (ModelIn{i1}.PTable(i3,1)*beta(ModelIn{i1}.PTable(i3,2),t+1)*bb(ModelIn{i1}.PTable(i3,2),t+1));
      end;
    end;
  end;
  beta(:,t) = beta(:,t) / rho(t+1);
end;

gamma=alpha.*beta;


% END OF FILE